#!/usr/bin/perl -w

# Program apmparser.pl

# (c) Gary C. Kessler, 2011 [gck@garykessler.net]

# Program to parse Apple Partition Map data structures.
# Ref: File System Forensic Analysis, Brian Carrier
#  and http://www.informit.com/articles/article.aspx?p=376123&seqNum=3

# General structure of the APM
#
# SECTOR 0: Legacy (or Protective) MBR
# SECTORS 2-X: One partition map/sector

# ********************************************************
# MAIN PROGRAM BLOCK
# ********************************************************

#use Encode qw/encode_utf8 decode_utf8/;
use utf8;
binmode STDOUT, ":utf8";

# Initialize global variables

$version = "1.5";
$build_date = "14 March 2013";

# $source is the name of the file to read.

$source = "";

%partition_status =
  (
  "0x00000001"=>"Entry is valid (A/UX only)",
  "0x00000002"=>"Entry is allocated (A/UX only)",
  "0x00000004"=>"Entry in use (A/UX only)",
  "0x00000008"=>"Entry contains boot information (A/UX only)",
  "0x00000010"=>"Partition is readable (A/UX only)",
  "0x00000020"=>"Partition is writable (Macintosh & A/UX)",
  "0x00000040"=>"Boot code is position independent (A/UX only)",
  "0x00000100"=>"Partition contains chain-compatible driver (Macintosh only)",
  "0x00000200"=>"Partition contains a real driver (Macintosh only)",
  "0x00000400"=>"Partition contains a chain driver (Macintosh only)",
  "0x40000000"=>"Automatically mount at startup (Macintosh only)",
  "0x80000000"=>"The startup partition (Macintosh only)"
  );

#
# Begin program activity!!
#

print "\nAPM Parser V$version - Gary C. Kessler ($build_date)\n\n";

# Parse command for file name and request missing information.

if (parsecommandline ())
  { exit 1; };

print "Source file = $source\n";

# Open input ($source) file and stop program if error

open (INFILE, "<:raw", $source)
  or die "Can't open input file \"$source\": $!";
binmode INFILE;

# Start by reading the entire file. Set the same max as for GPT headers, namely 17,408 bytes
# (34 sectors) in length

$header_len = 0;
while (not (eof INFILE) && $header_len < 17408)
  {
  read (INFILE,$byte,1);
  $header [$header_len] = unpack ("C", $byte);
  $header_len++;
  }

close INFILE;
printf ("Input file length = %d bytes (%d sectors).\n\n", $header_len, $header_len/512);

if ($header_len < 512)
  { die "File is smaller than one sector ($len bytes)\n"; }
if ($header_len%512 != 0)
  { die "File is not an integer number of sectors ($len bytes)\n"; }

# File may start with a sector 0 which is not processed.
# Determine which and process accordingly. (Assume sector 1 if the header
# starts with "PM"

# $sector is the sector number that is currently being processed
# $bbase is the absolute byte position in the header of byte 0 in the sector
#   currently being processed
# $bpos is the byte position in the sector; reset to 0 with each new sector

# $bpos, then is the relative byte position within the LBA, $bbase+$bpos is
#    the absolute byte position within $header

if ($header [0] == 0x50 && $header [1] == 0x4D)
    { 
    $first_sector = 1;
    $bbase = -512;
    }
  else

# If sector 0 is present, merely display it.

    {
    $first_sector = 0;
    $bbase = 0;
    print "***** Sector 0: (Not processed) *****\n\n";
    for ($i=0; $i<32; $i++)
      {
      $bpos = $i * 16;
      set_output ($bpos,16,@header);
      }
    print "\n";
    }

# Process remaining sectors, treating each as a partition table entry
# NOTE: Values here are big endian!!

# num_sectors is the number of sectors to process in the APM

$num_sectors = ($header_len / 512) - (1 - $first_sector);

for ($sector = 1; $sector <= $num_sectors; $sector++)
  {
  $bbase += 512;
  print "***** Sector $sector: Partition Table Header *****\n\n";

# Bytes 0-1 contain the signature (0x504D)

  $bpos = 0; $len = 2;
  printf ("%03d-%03d  Signature value: ", $bpos,$bpos+$len-1);
  list_hex_digits ($bbase+$bpos, $len, @header);
  print " [";
  $part_sig = "";
  for ($i = $bpos; $i < $bpos+$len; $i++)
    {
    print chr($header[$bbase+$bpos+$i]);
    $part_sig = $part_sig . chr($header[$bbase+$bpos+$i]);
    }

  if ($part_sig eq "PM")
      { print "] (Valid signature)\n"; }
    else
      { print "] (Invalid signature)\n"; }

# Bytes 2-3 are reserved and should be 0

  $bpos = 2; $len = 2;
  printf ("%03d-%03d  (Reserved): ", $bpos,$bpos+$len-1);
  list_hex_digits ($bbase+$bpos, $len, @header);
  print "\n";

# Bytes 4-7 contain the total number of partitions

  $bpos = 4; $len = 4;
  printf ("%03d-%03d  Total number of partitions: ", $bpos,$bpos+$len-1);
  list_hex_digits ($bbase+$bpos, $len, @header);
  printf (" [%s]\n", commify (big_endian ($bbase+$bpos,4,@header)));

# Bytes 8-11 contain the starting sector of the partition

  $bpos = 8; $len = 4;
  printf ("%03d-%03d  Starting sector of partition: ", $bpos,$bpos+$len-1);
  list_hex_digits ($bbase+$bpos, $len, @header);
  printf (" [%s]\n", commify(big_endian ($bbase+$bpos,4,@header)));

# Bytes 12-15 contain the partition size (sectors)

  $bpos = 12; $len = 4;
  printf ("%03d-%03d  Partition size: ", $bpos,$bpos+$len-1);
  list_hex_digits ($bbase+$bpos, $len, @header);
  printf (" [%s sectors]\n", commify(big_endian ($bbase+$bpos,4,@header)));

# Bytes 16-47 contain the partition name

  $bpos = 16; $len = 32;
  printf ("%03d-%03d  Partition name: \n", $bpos,$bpos+$len-1);

  for ($i = $bpos; $i < $bpos+$len; $i++)
    { $temp [$i] = $header [$bbase+$i]; }

  for ($i=0; $i<2; $i++)
    {
    $bpos = $i * 16 + 16;
    set_output ($bpos,16,@temp);
    }

# Bytes 48-79 contain the partition type

  $bpos = 48; $len = 32;
  printf ("%03d-%03d  Partition type: \n", $bpos,$bpos+$len-1);

  for ($i = $bpos; $i < $bpos+$len; $i++)
    { $temp [$i] = $header [$bbase+$i]; }

  for ($i=0; $i<2; $i++)
    {
    $bpos = $i * 16 + 48;
    set_output ($bpos,16,@temp);
    }

# Bytes 80-83 contain the starting sector of the data area

  $bpos = 80; $len = 4;
  printf ("%03d-%03d  Starting sector of data area: ", $bpos,$bpos+$len-1);
  list_hex_digits ($bbase+$bpos, $len, @header);
  printf (" [%s]\n", commify (big_endian ($bbase+$bpos,4,@header)));

# Bytes 84-87 contain the size of data area (sectors)

  $bpos = 84; $len = 4;
  printf ("%03d-%03d  Data area size: ", $bpos,$bpos+$len-1);
  list_hex_digits ($bbase+$bpos, $len, @header);
  printf (" [%s sectors]\n", commify(big_endian ($bbase+$bpos,4,@header)));

# Bytes 88-91 contain the partition status

  $bpos = 88; $len = 4;
  printf ("%03d-%03d  Partition status: ", $bpos,$bpos+$len-1);
  list_hex_digits ($bbase+$bpos, $len, @header);
  print "\n";
  $temp = big_endian ($bbase+$bpos,4,@header);
  for ($i=0; $i<32; $i++)
    {
    if ($temp%2 == 1)
      {
      printf ("         Bit %2d: ", $i);
      $flag = sprintf ("0x%08X", 2**$i);
      if (defined $partition_status{$flag})
          { print "$partition_status{$flag}\n"; }
        else
          { print "(Not listed)\n"; }
       }
     $temp = int ($temp / 2);
     }

# Bytes 92-95 contain the starting sector of the boot code

  $bpos = 92; $len = 4;
  printf ("%03d-%03d  Starting sector of boot code: ", $bpos,$bpos+$len-1);
  list_hex_digits ($bbase+$bpos, $len, @header);
  printf (" [%s]\n", commify (big_endian ($bbase+$bpos,4,@header)));

# Bytes 96-99 contain the size of the boot code

  $bpos = 96; $len = 4;
  printf ("%03d-%03d  Size of boot code: ", $bpos,$bpos+$len-1);
  list_hex_digits ($bbase+$bpos, $len, @header);
  printf (" [%s sectors]\n", commify (big_endian ($bbase+$bpos,4,@header)));

# Bytes 100-103 contain the address of boot loader code

  $bpos = 100; $len = 4;
  printf ("%03d-%03d  Address of boot loader code: ", $bpos,$bpos+$len-1);
  list_hex_digits ($bbase+$bpos, $len, @header);
  printf (" [%s]\n", commify (big_endian ($bbase+$bpos,4,@header)));

# Bytes 104-107 are reserved

  $bpos = 104; $len = 4;
  printf ("%03d-%03d  (Reserved): ", $bpos,$bpos+$len-1);
  list_hex_digits ($bbase+$bpos, $len, @header);
  print "\n";

# Bytes 108-111 contain the boot code entry point

  $bpos = 108; $len = 4;
  printf ("%03d-%03d  Boot code entry point: ", $bpos,$bpos+$len-1);
  list_hex_digits ($bbase+$bpos, $len, @header);
  printf (" [%s]\n", commify (big_endian ($bbase+$bpos,4,@header)));

# Bytes 112-115 are reserved

  $bpos = 112; $len = 4;
  printf ("%03d-%03d  (Reserved): ", $bpos,$bpos+$len-1);
  list_hex_digits ($bbase+$bpos, $len, @header);
  print "\n";

# Bytes 116-119 are the boot code checksum

  $bpos = 116; $len = 4;
  printf ("%03d-%03d  Boot code checksum: ", $bpos,$bpos+$len-1);
  list_hex_digits ($bbase+$bpos, $len, @header);
  print "\n";

# Bytes 120-135 contain the processor type

  $bpos = 120; $len = 16;
  printf ("%03d-%03d  Processor type: ", $bpos,$bpos+$len-1);
  list_hex_digits ($bbase+$bpos, $len, @header);
  print "\n";

# Bytes 136-511 are reserved

  $bpos = 136; $len = 376;
  printf ("%03d-%03d  (Reserved): \n", $bpos,$bpos+$len-1);

  for ($i = $bpos; $i < $bpos+$len; $i++)
    { $temp [$i] = $header [$bbase+$i]; }

  for ($i=0; $i<23; $i++)
    {
    set_output ($bpos,16,@temp);
    $bpos += 16;
    }

  set_output ($bpos,8,@temp);
  print "\n";
  }

# Finish up!

print "\nDone!\n";


# ********************************************************
# *****           SUBROUTINE BLOCKS                  *****
# ********************************************************

# ********************************************************
# big_endian
# ********************************************************

# Given an array name, starting position, and number of digits,
# calculate the decimal value using big endian (MSB first)

sub big_endian
{
my ($start,$n,@array) = @_;
my ($i, $sum);

$sum = $array[$start+$n-1];
for ($i=$n; $i>1; $i--)
  { $sum += $array[$start+$n-$i] * (2**(($i-1)*8)); }

return $sum;
}

# ********************************************************
# commify
# ********************************************************

# Routine from PerlFAQ5 -- add commas to large integers
# http://perldoc.perl.org/perlfaq5.html
  #How-can-I-output-my-numbers-with-commas-added?

sub commify

{
local $_  = shift;
1 while s/^([-+]?\d+)(\d{3})/$1,$2/;
return $_;
}


# ********************************************************
# help_text
# ********************************************************

# Display the help file

sub help_text
{
print<< "EOT";
Program usage: apmparser [-i input_file]
               apmparser {[-h] | [-v] | [-t]}

 where: -i is the input file name
        -h prints this help file
        -v displays the program version number
        -t prints the APM template file

If no -i switch is provided, the program will ask for the
input file name.

The input file should contain the contents of an Apple Partition.
Map (APM) header. The file should be a binary file that is no more
than 34 sectors (17,408 bytes) in length. The best way to obtain
the APM is to export the file from your forensics software.

EOT
return;
}


# ********************************************************
# list_hex_digits
# ********************************************************

# Given an array name, starting position, and number of digits, print
# out a string of hex digits in the format 0xAA-BB-CC...

sub list_hex_digits
{
my ($start,$n,@array) = @_;
my $i;

for ($i = $start; $i < $start+$n; $i++)
  {
  if ($i == $start)
      { print "0x"; }
    else
      { print "-"; }
  printf ("%02X", $array[$i]);
  }

return;
}


# ********************************************************
# parsecommandline
# ********************************************************

# Parse command line for file name, if present. Query
# user for any missing information

# Return $state = 1 to indicate that the program should stop
# immediately (switch -h)

sub parsecommandline
{
my $state = 0;

# Parse command line switches ($ARGV array of length $#ARGV)

if ($#ARGV >= 0)
  { 
  for ($i = 0; $i <= $#ARGV; $i++)
    {
    PARMS:
      {
      $ARGV[$i] eq "-i" && do
         {
         $source = $ARGV[$i+1];
         $i++;
         last PARMS;
         };
      $ARGV[$i] eq "-t" && do
         {
         $layout = "layout_APM.txt";
         open (INFILE, "<", $layout)
          or die "Can't open input file \"$layout\": $!";
         while (not (eof INFILE))
           {
           $line = <INFILE>;
           print $line;
           }
         close INFILE;
         $state = 1;
         return $state;
         };
      $ARGV[$i] eq "-v" && do
         {
         $state = 1;
         return $state;
         };
      $ARGV[$i] eq "-h" && do
         {
         help_text();
         $state = 1;
         return $state;
         };

      do
         {
         print "Invalid parameter \"$ARGV[$i]\".\n\n";
         print "Usage: apmparser [-i input_file]\n";
         print "       apmparser {[-h] | [-v] | [-t]}\n\n";
         $state = 1;
         return $state;
         };
      };
    };
  };

# Prompt for file names, if not supplied

if ($source eq "")
  {
  print "Enter the input file name: ";
  chomp ($source = <STDIN>);
  print "\n";
  };

return $state;
}


# ********************************************************
# set_output
# ********************************************************

# Print output of the data bytes to show both hex and ASCII
# values.

sub set_output
{
my ($bpos,$n,@array) = @_;
my ($j, $k, $linemax);

# Format for a maximum of 16 items per line

$linemax = 16;

printf ("%03d: ", $bpos);
for ($j=0; $j<$n; $j++)
  {
  printf (" %02X", $array[$bpos+$j]);
  }
if ($n < $linemax)
  {
  for ($j=1; $j<=$linemax-$n; $j++)
    { print "   "; };
  }

print "   ";
for ($j=0; $j<$n; $j++)
  {
  $k = $array[$bpos+$j];

#   Print a dot (.) for unprintable characters

  if ($k <= 0x1f || ($k >= 0x7f && $k <= 0x9f))
      { print "."; }
    else
      { print chr($k); }
  }
print "\n";

return;
}


